# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os

from servicex.transformer.object_store_manager import ObjectStoreManager


class TestObjectStoreManager:
    def test_init(self, mocker):
        mock_minio = mocker.patch('minio.Minio')
        ObjectStoreManager('localhost:9999', 'foo', 'bar')
        called_config = mock_minio.call_args[1]
        assert called_config['endpoint'] == 'localhost:9999'
        assert called_config['access_key'] == 'foo'
        assert called_config['secret_key'] == 'bar'
        assert not called_config['secure']

    def test_init_from_env(self, mocker):
        os.environ['MINIO_URL'] = 'localhost:9999'
        os.environ['MINIO_ACCESS_KEY'] = 'test'
        os.environ['MINIO_SECRET_KEY'] = 'shhh'
        mock_minio = mocker.patch('minio.Minio')

        ObjectStoreManager()
        called_config = mock_minio.call_args[1]
        assert called_config['endpoint'] == 'localhost:9999'
        assert called_config['access_key'] == 'test'
        assert called_config['secret_key'] == 'shhh'
        assert not called_config['secure']

    def test_upload_file(self, mocker):
        import minio
        mock_minio = mocker.MagicMock(minio.api.Minio)
        mock_minio.fput_object = mocker.Mock()
        mocker.patch('minio.Minio', return_value=mock_minio)
        result = ObjectStoreManager('localhost:9999', 'foo', 'bar')
        result.upload_file("my-bucket", "foo.txt", "/tmp/foo.txt")
        mock_minio.fput_object.assert_called()



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import pika

from servicex.transformer.rabbit_mq_manager import RabbitMQManager


class TestRabbitMQManager:
    def test_init(self, mocker):
        def callback():
            return "hi"

        mock_url_parameters = mocker.patch('pika.URLParameters', return_value="mock_url")
        mock_channel = mocker.MagicMock(pika.adapters.blocking_connection.BlockingChannel)
        mock_channel.basic_qos = mocker.Mock()
        mock_channel.basic_consume = mocker.Mock()
        mock_channel.start_consuming = mocker.Mock()
        mock_conn = mocker.MagicMock(pika.BlockingConnection)
        mock_conn.channel = mocker.Mock(return_value=mock_channel)

        mock_pika = mocker.patch('pika.BlockingConnection', return_value=mock_conn)

        RabbitMQManager("http:/foo.com", "servicex", callback)

        mock_url_parameters.assert_called_with("http:/foo.com")
        mock_pika.assert_called_with("mock_url")
        mock_channel.basic_qos.assert_called_with(prefetch_count=1)

        basic_consume_config = mock_channel.basic_consume.call_args[1]
        assert basic_consume_config['queue'] == 'servicex'
        assert not basic_consume_config['auto_ack']
        mock_channel.start_consuming.assert_called()



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import time
import pyarrow as pa


class ArrowWriter:

    def __init__(self, file_format=None, object_store=None, messaging=None):
        self.file_format = file_format
        self.object_store = object_store
        self.messaging = messaging
        self.messaging_timings = []
        self.object_store_timing = 0

    def write_branches_to_arrow(self, transformer,
                                topic_name, file_id, request_id):
        from .scratch_file_writer import ScratchFileWriter

        tick = time.time()

        scratch_writer = None

        batch_number = 0
        total_events = 0
        total_bytes = 0
        for pa_table in transformer.arrow_table():
            if self.object_store:
                if not scratch_writer:
                    scratch_writer = ScratchFileWriter(file_format=self.file_format)
                    scratch_writer.open_scratch_file(pa_table)

                scratch_writer.append_table_to_scratch(pa_table)

            total_events = total_events + pa_table.num_rows
            batches = pa_table.to_batches(max_chunksize=transformer.chunk_size)

            for batch in batches:
                if self.messaging:
                    messaging_tick = time.time()
                    key = str.encode(transformer.file_path + "-" + str(batch_number))

                    sink = pa.BufferOutputStream()
                    writer = pa.RecordBatchStreamWriter(sink, batch.schema)
                    writer.write_batch(batch)
                    writer.close()
                    self.messaging.publish_message(
                        topic_name,
                        key,
                        sink.getvalue())

                    total_bytes = total_bytes + len(sink.getvalue().to_pybytes())

                    avg_cell_size = len(sink.getvalue().to_pybytes()) / len(
                        transformer.attr_name_list) / batch.num_rows
                    print("Batch number " + str(batch_number) + ", "
                          + str(batch.num_rows) +
                          " events published to " + topic_name,
                          "Avg Cell Size = " + str(avg_cell_size) + " bytes")
                    batch_number += 1
                    self.messaging_timings.append(time.time() - messaging_tick)

                    # if server_endpoint:
                    #     post_status_update(server_endpoint, "Processed " +
                    #                        str(batch.num_rows))

        if self.object_store:
            object_store_tick = time.time()
            scratch_writer.close_scratch_file()

            print("Writing parquet to ", request_id, " as ",
                  transformer.file_path.replace('/', ':'))

            self.object_store.upload_file(request_id,
                                          transformer.file_path.replace('/', ':'),
                                          scratch_writer.file_path)

            scratch_writer.remove_scratch_file()
            self.object_store_timing = time.time() - object_store_tick

        tock = time.time()
        print("Real time: " + str(round(tock - tick / 60.0, 2)) + " minutes")



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from collections import OrderedDict

from servicex.transformer.arrow_writer import ArrowWriter
from servicex.transformer.object_store_manager import ObjectStoreManager
from servicex.transformer.kafka_messaging import KafkaMessaging
from servicex.transformer.uproot_transformer import UprootTransformer
import pyarrow as pa

# Mock module moved in python 3.3
import sys
if sys.version_info >= (3, 3):
    from unittest.mock import call  # noqa: E402
else:
    from mock import call  # noqa: E402


class TestArrowWriter:
    def test_init(self, mocker):
        mock_object_store = mocker.MagicMock(ObjectStoreManager)
        mock_messaging = mocker.MagicMock(KafkaMessaging)

        aw = ArrowWriter(file_format='hdf5',
                         object_store=mock_object_store,
                         messaging=mock_messaging)

        assert aw.object_store == mock_object_store
        assert aw.file_format == 'hdf5'
        assert aw.messaging == mock_messaging

    def test_transform_file_object_store(self, mocker):
        from servicex.transformer.scratch_file_writer import ScratchFileWriter

        mock_object_store = mocker.MagicMock(ObjectStoreManager)
        mock_scratch_file = mocker.MagicMock(ScratchFileWriter)
        scratch_file_init = mocker.patch(
            "servicex.transformer.scratch_file_writer.ScratchFileWriter",
            return_value=mock_scratch_file)

        mock_scratch_file.file_path = "/tmp/foo"

        aw = ArrowWriter(file_format='parquet',
                         object_store=mock_object_store,
                         messaging=None)

        mock_transformer = mocker.MagicMock(UprootTransformer)
        mock_transformer.file_path = '/tmp/foo'
        mock_transformer.chunk_size = 100
        mock_transformer.attr_name_list = ['a', 'b']

        data = OrderedDict([('strs', [chr(c) for c in range(ord('a'), ord('n'))]),
                            ('ints', list(range(1, 14)))])
        table = pa.Table.from_pydict(data)
        table2 = pa.Table.from_pydict(data)

        mock_transformer.arrow_table = mocker.Mock(return_value=iter([table, table2]))
        aw.write_branches_to_arrow(transformer=mock_transformer, topic_name='servicex',
                                   file_id=42, request_id="123-45")

        scratch_file_init.assert_called_with(file_format='parquet')
        mock_transformer.arrow_table.assert_called_with()
        mock_scratch_file.open_scratch_file.assert_called_once_with(table)
        mock_scratch_file.append_table_to_scratch.assert_has_calls(
            [call(table), call(table2)])

        mock_scratch_file.close_scratch_file.assert_called_once()

        mock_object_store.upload_file.assert_called_once_with("123-45", ":tmp:foo",
                                                              "/tmp/foo")
        mock_scratch_file.remove_scratch_file.assert_called_once()

    def test_transform_file_kafka(self, mocker):
        mock_kafka = mocker.MagicMock(KafkaMessaging)

        aw = ArrowWriter(file_format='parquet',
                         object_store=None,
                         messaging=mock_kafka)

        mock_transformer = mocker.MagicMock(UprootTransformer)
        mock_transformer.file_path = '/tmp/foo'
        mock_transformer.chunk_size = 100
        mock_transformer.attr_name_list = ['a', 'b']

        data = OrderedDict([('strs', [chr(c) for c in range(ord('a'), ord('n'))]),
                            ('ints', list(range(1, 14)))])
        table = pa.Table.from_pydict(data)
        table2 = pa.Table.from_pydict(data)

        mock_transformer.arrow_table = mocker.Mock(return_value=iter([table, table2]))
        aw.write_branches_to_arrow(transformer=mock_transformer, topic_name='servicex',
                                   file_id=42, request_id="123-45")

        mock_transformer.arrow_table.assert_called_with()

        kafka_calls = mock_kafka.publish_message.call_args_list
        assert len(kafka_calls) == 2
        topic, key, py_arrow_buffer = kafka_calls[0][0]
        assert topic == 'servicex'
        assert key == b'/tmp/foo-0'

        reader = pa.RecordBatchStreamReader(py_arrow_buffer)
        table = reader.read_all()
        assert table.column_names == ['strs', 'ints']



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from servicex.transformer.servicex_adapter import ServiceXAdapter


class TestServiceXAdapter:

    def test_init(self, mocker):
        import requests
        mock_session = mocker.MagicMock(requests.session)
        mocker.patch('requests.session', return_value=mock_session)
        mock_session.mount = mocker.Mock()
        ServiceXAdapter("http://foo.com")
        retries = mock_session.mount.mock_calls[0][1][1].max_retries
        assert retries.total == 5
        assert retries.connect == 3

    def test_put_file_complete(self, mocker):
        import requests

        mock_session = mocker.MagicMock(requests.session)
        mock_session.mount = mocker.Mock()
        mock_session.put = mocker.Mock()
        mocker.patch('requests.session', return_value=mock_session)

        adapter = ServiceXAdapter("http://foo.com")
        adapter.put_file_complete("my-root.root", 42, "testing", 1, 2, 3, 4)
        mock_session.put.assert_called()
        args = mock_session.put.call_args
        assert args[0][0] == 'http://foo.com/file-complete'
        doc = args[1]['json']
        assert doc['status'] == 'testing'
        assert doc['total-events'] == 3
        assert doc['total-time'] == 2
        assert doc['file-path'] == 'my-root.root'
        assert doc['num-messages'] == 1
        assert doc['file-id'] == 42
        assert doc['avg-rate'] == 1

    def test_put_file_complete_retry(self, mocker):
        import requests

        mock_session = mocker.MagicMock(requests.session)
        mock_session.mount = mocker.Mock()
        mock_session.put = mocker.Mock(side_effect=[requests.exceptions.ConnectionError, 200])
        mocker.patch('requests.session', return_value=mock_session)

        adapter = ServiceXAdapter("http://foo.com")
        adapter.put_file_complete("my-root.root", 42, "testing", 1, 2, 3, 4)
        assert mock_session.put.call_count == 2

    def test_post_status_update(self, mocker):
        import requests
        import os
        mocker.patch.dict(os.environ, {"POD_NAME": "my-pod"})

        mock_session = mocker.MagicMock(requests.session)
        mock_session.mount = mocker.Mock()
        mock_session.post = mocker.Mock()
        mocker.patch('requests.session', return_value=mock_session)

        adapter = ServiceXAdapter("http://foo.com")
        adapter.post_status_update(42, "testing", "this is a test")
        mock_session.post.assert_called()
        args = mock_session.post.call_args
        assert args[0][0] == 'http://foo.com/42/status'
        doc = args[1]['data']
        assert doc['status-code'] == 'testing'
        assert doc['info'] == 'this is a test'
        assert doc['pod-name'] == 'my-pod'

    def test_post_status_update_retry(self, mocker):
        import requests
        import os
        mocker.patch.dict(os.environ, {"POD_NAME": "my-pod"})

        mock_session = mocker.MagicMock(requests.session)
        mock_session.mount = mocker.Mock()
        mock_session.post = mocker.Mock(
            side_effect=[requests.exceptions.ConnectionError, 200])
        mocker.patch('requests.session', return_value=mock_session)

        adapter = ServiceXAdapter("http://foo.com")
        adapter.post_status_update(42, "testing", "this is a test")
        assert mock_session.post.call_count == 2
